/*------------------------------------------------------*/
/* A "life"-like automaton in 3 dimensions, with	*/
/* per-cell rulesets that compete with each other.	*/
/*------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <values.h>
#include <X11/Intrinsic.h>

#include <tcl.h>

#include "defines.h"

/*------------------------------------------------------*/
/* Global variables exported by other source files	*/
/*------------------------------------------------------*/

extern Dimension width, height;	   /* viewport boundaries (x=0, y=0 is lower bound) */

/*------------------------------------------------------*/
/* Global variables used by the simulator		*/
/*------------------------------------------------------*/

Tcl_HashTable TypeHashTable;
Tcl_HashTable CellHashTable;
Tcl_HashTable DNARuleSets;
Tcl_HashTable CheckEventTable;
ChangeRecord *ChangeEventStack;

Type *SpaceType;
int verbose;
int zplane;
char zslice;
unsigned char *typetable;
unsigned int num_types;

/*------------------------------*/
/* Random number initialization */
/*------------------------------*/

void setrandom()
{
   struct timeval ltv;
   struct timezone ltz;

   gettimeofday(&ltv, &ltz);

   srandom((int)ltv.tv_usec);
}

/*--------------------------------------------------------------*/
/* Generate a string from a Point3D to use as a hash key	*/
/*--------------------------------------------------------------*/

char *PointToHash(Point3D *point) {
   static char key[19];

   key[0] = 32 +  (point->x & 0x0000003f);
   key[1] = 32 + ((point->x & 0x00000fc0) >> 6);
   key[2] = 32 + ((point->x & 0x0003f000) >> 12);
   key[3] = 32 + ((point->x & 0x00fc0000) >> 18);
   key[4] = 32 + ((point->x & 0x3f000000) >> 24);
   key[5] = 32 + ((point->x & 0xc0000000) >> 30);

   key[6] = 32 +  (point->y & 0x0000003f);
   key[7] = 32 + ((point->y & 0x00000fc0) >> 6);
   key[8] = 32 + ((point->y & 0x0003f000) >> 12);
   key[9] = 32 + ((point->y & 0x00fc0000) >> 18);
   key[10] = 32 + ((point->y & 0x3f000000) >> 24);
   key[11] = 32 + ((point->y & 0xc0000000) >> 30);

   key[12] = 32 +  (point->z & 0x0000003f);
   key[13] = 32 + ((point->z & 0x00000fc0) >> 6);
   key[14] = 32 + ((point->z & 0x0003f000) >> 12);
   key[15] = 32 + ((point->z & 0x00fc0000) >> 18);
   key[16] = 32 + ((point->z & 0x3f000000) >> 24);
   key[17] = 32 + ((point->z & 0xc0000000) >> 30);

   key[18] = 0;

   return (char *)key;
}

/*--------------------------------------------------------------*/
/* Inverted function, get Point3D from hash value		*/
/*--------------------------------------------------------------*/

Point3D *HashToPoint(char *phash) {
   static Point3D point;

   point.x = phash[0] - 32;
   point.x |= (phash[1] - 32) << 6;
   point.x |= (phash[2] - 32) << 12;
   point.x |= (phash[3] - 32) << 18;
   point.x |= (phash[4] - 32) << 24;
   point.x |= (phash[5] - 32) << 30;

   point.y = phash[6] - 32;
   point.y |= (phash[7] - 32) << 6;
   point.y |= (phash[8] - 32) << 12;
   point.y |= (phash[9] - 32) << 18;
   point.y |= (phash[10] - 32) << 24;
   point.y |= (phash[11] - 32) << 30;

   point.z = phash[12] - 32;
   point.z |= (phash[13] - 32) << 6;
   point.z |= (phash[14] - 32) << 12;
   point.z |= (phash[15] - 32) << 18;
   point.z |= (phash[16] - 32) << 24;
   point.z |= (phash[17] - 32) << 30;
   
   return (Point3D *)(&point);
}

/*--------------------------------------------------------------*/
/* Replace the topmost change record				*/
/*--------------------------------------------------------------*/

void ReplaceChangeEvent(Type *type, DNARuleset *ruleset)
{
   ChangeRecord *event = ChangeEventStack;

   event->type = type;
   event->ruleset = ruleset;
}

/*--------------------------------------------------------------*/
/* Create and add a change event				*/
/*--------------------------------------------------------------*/

void AddChangeEvent(char *phash, Type *type, DNARuleset *ruleset)
{
   ChangeRecord *newevent;

   newevent = (ChangeRecord *)Tcl_Alloc(sizeof(ChangeRecord));
 
   strcpy(newevent->phash, phash);
   newevent->type = type;
   newevent->ruleset = ruleset;

   newevent->next = ChangeEventStack;
   ChangeEventStack = newevent;

   /* Diagnostic */
   if (verbose == 1) {
      Point3D *point = HashToPoint(phash);
      tcl_printf(stdout, "Change event at %d %d %d type 0x%x ruleset 0x%x\n",
		point->x, point->y, point->z, type, ruleset);
   }
}

/*--------------------------------------------------------------*/
/* Create and add a check event					*/
/*--------------------------------------------------------------*/

void AddCheckEvent(char *phash, DNARuleset *ruleset)
{
   CheckRecord *newevent, *checkevent;
   Tcl_HashEntry *he;
   int newptr;

   /* Search CheckEventTable for existing checks on this point. */
   /* Only check again if checking with a different ruleset.	*/

   he = Tcl_CreateHashEntry(&CheckEventTable, phash, &newptr);
   if (newptr == 1) {   

      newevent = (CheckRecord *)Tcl_Alloc(sizeof(CheckRecord));
      newevent->ruleset = ruleset;
      newevent->next = NULL;
      Tcl_SetHashValue(he, newevent);

      /* Diagnostic */
      if (verbose == 1) {
         Point3D *point = HashToPoint(phash);
	 tcl_printf(stdout, "Check event at %d %d %d ruleset 0x%x\n",
			point->x, point->y, point->z, ruleset);
      }
   }
   else {
      newevent = (CheckRecord *)Tcl_GetHashValue(he);
      checkevent = newevent;
      while ((checkevent->next != NULL) && (checkevent->ruleset != ruleset))
	 checkevent = checkevent->next;
      if (checkevent->ruleset != ruleset) {
	 // Set up a competing check for this space
         newevent = (CheckRecord *)Tcl_Alloc(sizeof(CheckRecord));
         newevent->ruleset = ruleset;
         newevent->next = NULL;
	 checkevent->next = newevent;

         /* Diagnostic */
	 if (verbose == 1) {
            Point3D *point = HashToPoint(phash);
            tcl_printf(stdout, "Competing Check event at %d %d %d ruleset 0x%x\n",
			point->x, point->y, point->z, ruleset);
	 }
      }
   }
}

/*--------------------------------------------------------------*/
/* Simulation step:  First process all change events, then	*/
/* process all check events.					*/
/*--------------------------------------------------------------*/

int simulate_step()
{
   Tcl_HashEntry *he, *he2, *he3, *he4;
   Tcl_HashSearch hs;
   Cell *cell, *testcell;
   ChangeRecord *change;
   CheckRecord *check, *checktop;
   int newptr, i;
   char needcheck;
   DNARule *rule;
   DNACond *cond;
   Type *type, *testtype;
   Point3D *point, testpt;
   unsigned int idx;

   // Offsets defining the 12 neighboring positions if we warp
   // a dense-packed hegaxonal volume to integer euclidean space.
   static char xoffsets[12] = {2, 1, -1, -2, -1,  1,  0, -1,  1,  0, -1,  1};
   static char yoffsets[12] = {0, 2,  2,  0, -2, -2, -1,  1,  1,  1, -1, -1};
   static char zoffsets[12] = {0, 0,  0,  0,  0,  0,  1,  1,  1, -1, -1, -1};

   /* Initialize the check event hash table */
   Tcl_InitHashTable(&CheckEventTable, TCL_STRING_KEYS);

   /* Remove each pending event from the current change event stack */

   if (verbose == 3)
      tcl_printf(stdout, "Processing change records\n");

   while (ChangeEventStack != NULL) {

      change = ChangeEventStack;
      ChangeEventStack = ChangeEventStack->next;
      needcheck = 0;

      // Change the type of a cell

      he = Tcl_FindHashEntry(&CellHashTable, change->phash);
      if (he == NULL) {
	 // Create a new cell
	 if (change->type != SpaceType) {
	    he = Tcl_CreateHashEntry(&CellHashTable, change->phash, &newptr);
	    cell = (Cell *)Tcl_Alloc(sizeof(Cell));
	    cell->type = change->type;
	    cell->ruleset = change->ruleset;
	    Tcl_SetHashValue(he, cell);
	    needcheck = 1;
	 }
      }
      else {
	 // Replace a cell
	 cell = (Cell *)Tcl_GetHashValue(he);
	 if (cell->type != change->type || cell->ruleset != change->ruleset) {
  	    if (change->type != SpaceType) {
	       cell->type = change->type;
	       cell->ruleset = change->ruleset;
	    }
	    else {
	       // Type becomes SpaceType, so delete the cell
	       Tcl_Free((char *)cell);
	       Tcl_DeleteHashEntry(he);
	    }
	    needcheck = 1;
	 }
      }

      if (needcheck) {

	 // Set all 12 neighboring positions to be checked on next step
	 point = HashToPoint(change->phash);
	 for (i = 0; i < 12; i++) {
	    testpt.x = point->x + xoffsets[i];
	    testpt.y = point->y + yoffsets[i];
	    testpt.z = point->z + zoffsets[i];
	    AddCheckEvent(PointToHash(&testpt), change->ruleset);
	 }
      }
      Tcl_Free((char *)change);
   }

   // Generate index into types, if it hasn't been made yet
   if (typetable == NULL)
      typetable = (unsigned char *)Tcl_Alloc(num_types * sizeof(unsigned char));

   /* Process the check table entries */

   if (verbose == 3) {
      char *statstr;
      tcl_printf(stdout, "Processing check records\n");
      statstr = Tcl_HashStats(&CheckEventTable);
      tcl_printf(stdout, "%s\n", statstr);
      Tcl_Free(statstr);
   }
      
   he = Tcl_FirstHashEntry(&CheckEventTable, &hs);
   while (he != NULL) {
      char *phash, *chash;
      char conflict = (char)0;
      unsigned int maxscore[2], score[2];

      check = (CheckRecord *)Tcl_GetHashValue(he);
      phash = Tcl_GetHashKey(&CheckEventTable, he);
      point = HashToPoint(phash);

      // Check if cell already exists at this point, get its type.
      // Otherwise, type is "space"

      he2 = Tcl_FindHashEntry(&CellHashTable, phash);
      if (he2 == NULL)
	 type = SpaceType;
      else {
	 cell = (Cell *)Tcl_GetHashValue(he2);
	 type = cell->type;
      }

      needcheck = 1;
      maxscore[0] = maxscore[1] = 0;

      // If there are conflicting rulesets being applied to this location,
      // there will be more than one entry to the check record.

      conflict = (change->next != NULL) ? (char)1 : (char)0;
      while (check != NULL) {

         // Evaluate the cell using the table passed in the event record

         he3 = Tcl_FindHashEntry(&(check->ruleset->DNARuleHash), (const char *)type);
         if (he3 != NULL) {   // NULL means no rules exist for this type

	    int j, matches;
	    unsigned int idx;

	    rule = Tcl_GetHashValue(he3);

	    // Use "needcheck" flag so we gather statistics the first time
	    // we have a valid rule for a type.  That way, we don't waste
	    // time gathering statistics for a non-existant rule.  

	    if (needcheck == 1) {
	       needcheck = 0;

	       // Clear previous counts
	       memset((void *)typetable, 0, (size_t)num_types);

	       // Collect statistics on the types surrounding this cell

	       for (i = 0; i < 12; i++) {
	          testpt.x = point->x + xoffsets[i];
	          testpt.y = point->y + yoffsets[i];
	          testpt.z = point->z + zoffsets[i];
	          chash = PointToHash(&testpt);
	 	  
	          he4 = Tcl_FindHashEntry(&CellHashTable, chash);
	          if (he4 == NULL)
	             testtype = SpaceType;
	          else {
	             testcell = (Cell *)Tcl_GetHashValue(he4);
	             testtype = testcell->type;
	          }
	          typetable[testtype->idx]++;
	       }		  
	    }
	
	    while (rule != NULL) {
	       // Evaluate rules, and apply a change/create order.
	       // Stop after 1st rule to evaluate true for all conditions.

	       score[0] = score[1] = 0;
	       cond = rule->condition;

	       while (cond != NULL) {
	          matches = 0;
	          idx = cond->type->idx;

	          switch(cond->operator) {
		     case EQUAL:
		        if (typetable[idx] == cond->number)
			   matches = 1;
		        break;
		     case NOT_EQUAL:
		        if (typetable[idx] != cond->number)
			   matches = 1;
		        break;
		     case LESS_THAN:
		        if (typetable[idx] < cond->number)
			   matches = 1;
		        break;
		     case MORE_THAN:
		        if (typetable[idx] > cond->number)
			   matches = 1;
		        break;
	          }
	          if (matches == 1) {
		     // First score is based on the number of cells
		     // involved in the rule.  Second score is based
		     // on the number of types involved in the rule.
		     score[0] += cond->number;
		     score[1]++;
		  }
		  else break;
	          cond = cond->next;
	       }
	       if (matches == 1) {	// We have a match.  Add event and stop
	          // But don't push results that don't change!
	          if (rule->result != type) {
		     if (maxscore[0] == 0) {
		        AddChangeEvent(phash, rule->result, check->ruleset);
			maxscore[0] = score[0];
			maxscore[1] = score[1];
		     }
		     else if (score[0] == maxscore[0]) {
			if (score[1] > maxscore[1]) {
		           ReplaceChangeEvent(rule->result, check->ruleset);
		           maxscore[1] = score[1];
			}
		     }
		     else if (score[0] > maxscore[0]) {
		        ReplaceChangeEvent(rule->result, check->ruleset);
		        maxscore[0] = score[0];
		        maxscore[1] = score[1];
		     }
	          }
	          break;
	       }
	       rule = rule->next;
	    }	
         }  

         // Be sure to remove the memory for this table entry

	 checktop = check;
	 check = check->next;
	 Tcl_Free((char *)checktop);
      }

      he = Tcl_NextHashEntry(&hs);
   }

   /* Remove the check hash table */
   Tcl_DeleteHashTable(&CheckEventTable);
}

/*--------------------------------------------------------------*/
/* Hash table destructors					*/
/*--------------------------------------------------------------*/

void destroyTypeHash()
{
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   Type *type;

   he = Tcl_FirstHashEntry(&TypeHashTable, &hs);
   while (he != NULL) {
      type = (Type *)Tcl_GetHashValue(he);
      Tcl_Free((char *)type);
      he = Tcl_NextHashEntry(&hs);
   }
   Tcl_DeleteHashTable(&TypeHashTable);

   Tcl_Free(typetable);
   typetable = NULL;
}

/*--------------------------------------------------------------*/

void destroyCellHash()
{
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   Cell *cell;

   he = Tcl_FirstHashEntry(&CellHashTable, &hs);
   while (he != NULL) {
      cell = (Cell *)Tcl_GetHashValue(he);
      Tcl_Free((char *)cell);
      he = Tcl_NextHashEntry(&hs);
   }
   Tcl_DeleteHashTable(&CellHashTable);
}

/*--------------------------------------------------------------*/

void destroyRuleSets()
{
   Tcl_HashSearch hs, hs2;
   Tcl_HashEntry *he, *he2;
   DNARuleset *ruleset;
   DNARule *rule, *nr;
   DNACond *cond, *nc;

   he = Tcl_FirstHashEntry(&DNARuleSets, &hs);
   while (he != NULL) {
      ruleset = (DNARuleset *)Tcl_GetHashValue(he);

      /* Each ruleset has a hash table that needs to be deleted */

      he2 = Tcl_FirstHashEntry(&(ruleset->DNARuleHash), &hs2);
      while (he2 != NULL) {
         rule = (DNARule *)Tcl_GetHashValue(he2);
	 while (rule != NULL) {
	    nr = rule->next;
	    cond = rule->condition;
	    while (cond != NULL) {
	       nc = cond->next;
	       Tcl_Free((char *)cond);
	       cond = nc;
	    }
	    Tcl_Free((char *)rule);
	    rule = nr;
	 }
         he2 = Tcl_NextHashEntry(&hs2);
      }
      Tcl_DeleteHashTable(&(ruleset->DNARuleHash));
      Tcl_Free((char *)ruleset);
      he = Tcl_NextHashEntry(&hs);
   }
   Tcl_DeleteHashTable(&DNARuleSets);
}

/*--------------------------------------------------------------*/
/* Reinitialization (at any time)				*/
/*--------------------------------------------------------------*/

void reinitialize(int option)
{
   ChangeEventStack = NULL;

   Tcl_InitHashTable(&CellHashTable, TCL_STRING_KEYS);
 
   if (option != 1) {
      Tcl_InitHashTable(&TypeHashTable, TCL_STRING_KEYS);
      Tcl_InitHashTable(&DNARuleSets, TCL_STRING_KEYS);
      Tcl_InitHashTable(&CheckEventTable, TCL_STRING_KEYS);

      typetable = NULL;
      num_types = 0;
      SpaceType = NULL;
   }

   /* Set graphics view angle back to default */
   viewangle[0] = 0.0;
   viewangle[1] = 0.0;
   viewangle[2] = 0.0;
   viewangle[3] = 0.0;
   viewzoom = 4.0;
}

/*--------------------------------------------------------------*/
/* Initialization at startup *only*				*/
/*--------------------------------------------------------------*/

void initialize()
{
   reinitialize(0);

   // Initial values

   verbose = 0;
   zslice = (char)0;
   zplane = 0;
   viewangle[0] = 0.0;
   viewangle[1] = 0.0;
   viewangle[2] = 0.0;
   viewangle[3] = 0.0;

   setrandom();
}

/*--------------------------------------------------------------*/
/* Destroy the world						*/
/*--------------------------------------------------------------*/

void destroy_world(int option)
{
   ChangeRecord *change;

   while (ChangeEventStack != NULL) {
      change = ChangeEventStack;
      ChangeEventStack = ChangeEventStack->next;
      Tcl_Free((char *)change);
   }
   destroyCellHash();

   if (!option) {
      destroyRuleSets();
      destroyTypeHash();

      Tcl_Free(typetable);
      typetable = NULL;
      num_types = 0;
      SpaceType = NULL;
   }
}

/*--------------------------------------------------------------*/
