/*	Graphics routines for OpenGL rendering			*/
/* Copyright (c) 2012  Tim Edwards, Open Circuit Design		*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include <tcl.h>
#include <tk.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "defines.h"

#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glu.h>

Display *dpy;
Colormap cmap;
Window   win;
Dimension width, height;
XFontStruct *textfont;

GLXContext      grXcontext;
GLuint		grXBases;
XVisualInfo     *grVisualInfo;

float		viewangle[4];
float		viewzoom;

/*--------------------------------------------------------------*/
/* Set line attributes						*/
/*--------------------------------------------------------------*/

void
SetLineAttributes(float lwidth)
{
   GLfloat glwidth = (lwidth <= 0) ? 1 : (GLfloat)lwidth;
   glLineWidth(glwidth);
   glEnable(GL_LINE_SMOOTH);
   glDisable(GL_LINE_STIPPLE);
}

/*--------------------------------------------------------------*/
/* Render the display into the draw buffer, then copy window	*/
/* up from buffer						*/
/*--------------------------------------------------------------*/

void draw_world()
{
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   Cell *cell;
   Point3D *point;
   char *phash;
   Type *type, *lasttype = NULL;
   GLfloat cvec[4];
   GLfloat light_pos[] = { 1.0, 1.0, 1.0, 1.0 };

   if (win == (Window)0) return;

   glXMakeCurrent(dpy, (GLXDrawable)win, grXcontext);
   glDrawBuffer(GL_BACK);
   glViewport((GLsizei)0, (GLsizei)0, width, height);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   /* Scale window so that 1 pixel = 1 point in simulation 	*/
   /* space at zoom factor 1.					*/
   /* Scale z very small, or depth buffer will overflow.	*/

   glScalef(1.0 / ((float)width / viewzoom),
   		-1.0 / ((float)height / viewzoom), 0.001);

   // glScalef(1.0 / (float)(width >> 2),
   //		-1.0 / (float)(height >> 2), 0.001);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glLightfv(GL_LIGHT0, GL_POSITION, light_pos);

   /* Set view angle */
   glRotatef((GLfloat)viewangle[0], (GLfloat)viewangle[1],
		(GLfloat)viewangle[2], (GLfloat)viewangle[3]);

   /* Scale to correct for difference between an integer grid	*/
   /* pretending to be hexagonal, and a true hexagonal grid.	*/
   glScalef(1.1547, 1.0, 2.0);

   glPointSize((GLfloat)(viewzoom * 1.2));
   glClearColor(1.0, 1.0, 1.0, 0.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // For z-plane slicing, I want to draw one plane opaque and the
   // rest translucent.  The technique for rendering translucent
   // objects in 3D is to render opaque objects first, make the
   // depth buffer read-only, then render transparent objects.
   // This requires two passes through the hash table.

   he = Tcl_FirstHashEntry(&CellHashTable, &hs);
   while (he != NULL) {
      phash = Tcl_GetHashKey(&CellHashTable, he);
      point = (Point3D *)HashToPoint(phash);

      if ((zslice == (char)0) || (point->z == zplane)) {

         cell = Tcl_GetHashValue(he);
         type = cell->type;

         if (type != lasttype) {
	    if (lasttype != NULL) glEnd();

	    cvec[0] = type->r;
	    cvec[1] = type->g;
	    cvec[2] = type->b;
	    cvec[3] = type->a;
	    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cvec);
	    glBegin(GL_POINTS);
	    lasttype = type;
	 }
         glVertex3i((GLint)point->x, (GLint)point->y, (GLint)point->z);
      }
      he = Tcl_NextHashEntry(&hs);
   }
   if (lasttype != NULL) glEnd();

   if (zslice == (char)1) {
      glDepthMask(GL_FALSE);

      lasttype = NULL;
      he = Tcl_FirstHashEntry(&CellHashTable, &hs);
      while (he != NULL) {
         phash = Tcl_GetHashKey(&CellHashTable, he);
         point = (Point3D *)HashToPoint(phash);

         if (point->z != zplane) {

            cell = Tcl_GetHashValue(he);
            type = cell->type;

            if (type != lasttype) {
	       if (lasttype != NULL) glEnd();

	       cvec[0] = type->r;
	       cvec[1] = type->g;
	       cvec[2] = type->b;
	       cvec[3] = 0.1;		/* Mostly transparent */
	       glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cvec);
	       glBegin(GL_POINTS);
	       lasttype = type;
	    }
            glVertex3i((GLint)point->x, (GLint)point->y, (GLint)point->z);
	 }
         he = Tcl_NextHashEntry(&hs);
      }
      if (lasttype != NULL) glEnd();
      glDepthMask(GL_TRUE);
   }

   glXSwapBuffers(dpy, (GLXDrawable)win);

   // Diagnostic
   if (verbose == 2) {
      he = Tcl_FirstHashEntry(&CellHashTable, &hs);
      while (he != NULL) {
         phash = Tcl_GetHashKey(&CellHashTable, he);
         point = (Point3D *)HashToPoint(phash);
         if ((zslice == (char)0) || (point->z == zplane)) {
            tcl_printf(stdout, "Draw type 0x%x at (%d %d %d)\n", type,
                        point->x, point->y, point->z);
	 }
      }
   }
}

/*--------------------------------------------------------------*/

u_char
LoadFont(Font font)
{
    grXBases = glGenLists(256);
    if (grXBases == 0)
       return 0;
    glXUseXFont(font, 0, 256, grXBases);
}

/*--------------------------------------------------------------*/

void resize(Tk_Window tkwind, int locwidth, int locheight)
{
   Window window;

   if ((locwidth == 0) || (locheight == 0)) return;

   if (Tk_WindowId(tkwind) == 0)
      Tk_MapWindow(tkwind);

   width = locwidth;
   height = locheight;
}

/*--------------------------------------------------------------*/

void expose(Tk_Window tkwind)
{
   if (Tk_WindowId(tkwind) == 0) return;
   draw_world();
}

/*--------------------------------------------------------------*/

int redraw(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   draw_world();
   return TCL_OK;
}

/*---------------------*/
/* Create a new window */
/*---------------------*/

int new_window(Tcl_Interp *interp, Tk_Window tkwind)
{
   char winpath[512];
   Font font;
   GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 0.0 };
   GLfloat shininess[] = { 50.0 };

   int attributeList[] = { GLX_RGBA, GLX_DOUBLEBUFFER, None };

   /* Before creating a new window structure, we need to make sure that	*/
   /* the target window is a "simple" window (from tkSimple.c), which 	*/
   /* we do by passing the command "<window_path> cget -commandproc"	*/
   /* and checking if it returns TCL_OK.				*/

   sprintf(winpath, "%s cget -commandproc", Tk_PathName(tkwind));
   if (Tcl_EvalEx(interp, winpath, -1, 0) != TCL_OK) {
      Tcl_SetResult(interp, "Window is the wrong type for rendering!", NULL);
      return -2;
   }

   /* Make sure the window is mapped */

   Tk_MapWindow(tkwind);
   dpy = Tk_Display(tkwind);
   win = Tk_WindowId(tkwind);
   width = Tk_Width(tkwind);
   height = Tk_Height(tkwind);
   cmap = Tk_Colormap(tkwind);

   grVisualInfo = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
   grXcontext = glXCreateContext(dpy, grVisualInfo, NULL, GL_FALSE);

   glXMakeCurrent(dpy, (GLXDrawable)win, grXcontext);

   /* All the things that need to be done only once */

   glEnable(GL_BLEND);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_POINT_SMOOTH);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glDisable(GL_COLOR_LOGIC_OP);
   SetLineAttributes(2.0);

   glShadeModel(GL_FLAT);

   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glDepthFunc(GL_LEQUAL);

   /* Fixed lighting properties */

   glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, shininess);

   /* Create the text font */

   font = XLoadFont(dpy, "-*-times-medium-r-normal--20-*-*-*-*-*-*-*");
   LoadFont(font);
   textfont = XQueryFont(dpy, (XID)font);

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   return 0;
}

/*--------------------------------------------------------------*/
/* GUI Initialization under Tk					*/
/* First argument is the Tk path name of the drawing window.	*/
/*--------------------------------------------------------------*/

int GUI_init(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Tk_Window 	tkwind, tktop;
   static char 	*dnadrawdefault = ".dna";
   char 	*dnadrawwin;
   int 		i, locobjc;
   XColor	scolor;

   int attributeList[] = { GLX_RGBA, GLX_DOUBLEBUFFER, None };

   tktop = Tk_MainWindow(interp);
   if (tktop == (Tk_Window)NULL) return TCL_ERROR;

   dnadrawwin = (char *)Tcl_GetVar(interp, "drawwindow", TCL_GLOBAL_ONLY);
   if (dnadrawwin == NULL)
      dnadrawwin = dnadrawdefault;

   tkwind = Tk_NameToWindow(interp, dnadrawwin, tktop);

   if (tkwind == NULL) {
      tcl_printf(stderr, "The Tk window hierarchy must be rooted at"
                ".dna or $drawwindow must point to the drawing window\n");
      return TCL_ERROR;
   }

   /* OpenGL setup */
   new_window(interp, tkwind);

   /* Initialize parameters for the simulation program */

   initialize();
   draw_world();

   return TCL_OK;
}

/*------------------------------------------------------*/
