/*--------------------------------------------------------------*/
/* tcldna.c:							*/
/*	Tcl routines for DNA command-line functions		*/
/* Copyright (c) 2012  Tim Edwards, Open Circuit Design, Inc.	*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <tk.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "defines.h"

/* Global variables */

Tcl_HashTable DNATagTable;
Tcl_Interp *dnainterp;
Tcl_Interp *consoleinterp;

/* Command structure */

typedef struct {
   const char	*cmdstr;
   void		(*func)();
} cmdstruct;

/* Forward declarations of commands */

extern int Tk_SimpleObjCmd();
extern int dna_start();
extern int redraw();
extern int single_step();
extern int QuitCallback();
extern int dnatcl_tag();
extern int InitCallback();
extern int dnatcl_verbose();
extern int dnatcl_zplane();
extern int dnatcl_cell();
extern int dnatcl_type();
extern int dnatcl_rule();
extern int dnatcl_pending();
extern int dnatcl_view();
extern int dnatcl_zoom();

static cmdstruct dna_commands[] =
{
   {"start", (void *)dna_start},
   {"redraw", (void *)redraw},
   {"step", (void *)single_step},
   {"quit", (void *)QuitCallback},
   {"tag", (void *)dnatcl_tag},
   {"init", (void *)InitCallback},
   {"verbose", (void *)dnatcl_verbose},
   {"cell", (void *)dnatcl_cell},
   {"type", (void *)dnatcl_type},
   {"rule", (void *)dnatcl_rule},
   {"pending", (void *)dnatcl_pending},
   {"zplane", (void *)dnatcl_zplane},
   {"view", (void *)dnatcl_view},
   {"zoom", (void *)dnatcl_zoom},
   {"", NULL}  /* sentinel */
};

/*-----------------------*/
/* Tcl 8.4 compatibility */
/*-----------------------*/

#ifndef CONST84
#define CONST84
#endif

/*----------------------------------------------------------------------*/
/* Reimplement strdup() to use Tcl_Alloc().				*/
/*----------------------------------------------------------------------*/

char *Tcl_Strdup(const char *s)
{
   char *snew;
   int slen;

   slen = 1 + strlen(s);
   snew = Tcl_Alloc(slen);
   if (snew != NULL)
      memcpy(snew, s, slen);

   return snew;
}

/*----------------------------------------------------------------------*/
/* Reimplement vfprintf() as a call to Tcl_Eval().			*/
/*----------------------------------------------------------------------*/

void tcl_vprintf(FILE *f, const char *fmt, va_list args_in)
{
   va_list args;
   static char outstr[128] = "puts -nonewline std";
   char *outptr, *bigstr = NULL, *finalstr = NULL;
   int i, nchars, result, escapes = 0, limit;

   /* If we are printing an error message, we want to bring attention	*/
   /* to it by mapping the console window and raising it, as necessary.	*/
   /* I'd rather do this internally than by Tcl_Eval(), but I can't	*/
   /* find the right window ID to map!					*/

   if ((f == stderr) && (consoleinterp != dnainterp)) {
      Tk_Window tkwind;
      tkwind = Tk_MainWindow(consoleinterp);
      if ((tkwind != NULL) && (!Tk_IsMapped(tkwind)))
	 result = Tcl_Eval(consoleinterp, "wm deiconify .\n");
      result = Tcl_Eval(consoleinterp, "raise .\n");
   }

   strcpy (outstr + 19, (f == stderr) ? "err \"" : "out \"");
   outptr = outstr;

   /* This mess circumvents problems with systems which do not have	*/
   /* va_copy() defined.  Some define __va_copy();  otherwise we must	*/
   /* assume that args = args_in is valid.				*/

   va_copy(args, args_in);
   nchars = vsnprintf(outptr + 24, 102, fmt, args);
   va_end(args);

   if (nchars >= 102) {
      va_copy(args, args_in);
      bigstr = Tcl_Alloc(nchars + 26);
      strncpy(bigstr, outptr, 24);
      outptr = bigstr;
      vsnprintf(outptr + 24, nchars + 2, fmt, args);
      va_end(args);
    }
    else if (nchars == -1) nchars = 126;

    for (i = 24; *(outptr + i) != '\0'; i++) {
       if (*(outptr + i) == '\"' || *(outptr + i) == '[' ||
	  	*(outptr + i) == ']' || *(outptr + i) == '\\')
	  escapes++;
    }

    if (escapes > 0) {
      finalstr = Tcl_Alloc(nchars + escapes + 26);
      strncpy(finalstr, outptr, 24);
      escapes = 0;
      for (i = 24; *(outptr + i) != '\0'; i++) {
	  if (*(outptr + i) == '\"' || *(outptr + i) == '[' ||
	    		*(outptr + i) == ']' || *(outptr + i) == '\\') {
	     *(finalstr + i + escapes) = '\\';
	     escapes++;
	  }
	  *(finalstr + i + escapes) = *(outptr + i);
      }
      outptr = finalstr;
    }

    *(outptr + 24 + nchars + escapes) = '\"';
    *(outptr + 25 + nchars + escapes) = '\0';

    result = Tcl_Eval(consoleinterp, outptr);

    if (bigstr != NULL) Tcl_Free(bigstr);
    if (finalstr != NULL) Tcl_Free(finalstr);
}
    
/*------------------------------------------------------*/
/* Console output flushing which goes along with the	*/
/* routine tcl_vprintf() above.				*/
/*------------------------------------------------------*/

void tcl_stdflush(FILE *f)
{   
   Tcl_SavedResult state;
   static char stdstr[] = "::flush stdxxx";
   char *stdptr = stdstr + 11;
    
   Tcl_SaveResult(dnainterp, &state);
   strcpy(stdptr, (f == stderr) ? "err" : "out");
   Tcl_Eval(dnainterp, stdstr);
   Tcl_RestoreResult(dnainterp, &state);
}

/*----------------------------------------------------------------------*/
/* Reimplement fprintf() as a call to Tcl_Eval().			*/
/*----------------------------------------------------------------------*/

void tcl_printf(FILE *f, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  tcl_vprintf(f, format, ap);
  va_end(ap);
}

/*----------------------------------------------------------------------*/
/* Implement tag callbacks on functions					*/
/* Find any tags associated with a command and execute them.		*/
/*----------------------------------------------------------------------*/

int DNATagCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int objidx, result = TCL_OK;
    char *postcmd, *substcmd, *newcmd, *sptr, *sres;
    char *croot = Tcl_GetString(objv[0]);
    Tcl_HashEntry *entry;
    Tcl_SavedResult state;
    int reset = FALSE;
    int i, llen, cmdnum;

    entry = Tcl_FindHashEntry(&DNATagTable, croot);
    postcmd = (entry) ? (char *)Tcl_GetHashValue(entry) : NULL;

    if (postcmd)
    {
	substcmd = (char *)Tcl_Alloc(strlen(postcmd) + 1);
	strcpy(substcmd, postcmd);
	sptr = substcmd;

	/*--------------------------------------------------------------*/
	/* Parse "postcmd" for Tk-substitution escapes			*/
	/* Allowed escapes are:						*/
	/* 	%W	substitute the tk path of the calling window	*/
	/*	%r	substitute the previous Tcl result string	*/
	/*	%R	substitute the previous Tcl result string and	*/
	/*		reset the Tcl result.				*/
	/*	%[0-5]  substitute the argument to the original command	*/
	/*	%N	substitute all arguments as a list		*/
	/*	%%	substitute a single percent character		*/
	/*	%*	(all others) no action: print as-is.		*/
	/*--------------------------------------------------------------*/

	while ((sptr = strchr(sptr, '%')) != NULL)
	{
	    switch (*(sptr + 1))
	    {
		case 'W': {
		    char *tkpath = NULL;
		    Tk_Window tkwind = Tk_MainWindow(interp);
		    if (tkwind != NULL) tkpath = Tk_PathName(tkwind);
		    if (tkpath == NULL)
			newcmd = (char *)Tcl_Alloc(strlen(substcmd));
		    else
			newcmd = (char *)Tcl_Alloc(strlen(substcmd) + strlen(tkpath));

		    strcpy(newcmd, substcmd);

		    if (tkpath == NULL)
			strcpy(newcmd + (int)(sptr - substcmd), sptr + 2);
		    else
		    {
			strcpy(newcmd + (int)(sptr - substcmd), tkpath);
			strcat(newcmd, sptr + 2);
		    }
		    Tcl_Free(substcmd);
		    substcmd = newcmd;
		    sptr = substcmd;
		    } break;

		case 'R':
		    reset = TRUE;
		case 'r':
		    sres = (char *)Tcl_GetStringResult(interp);
		    newcmd = (char *)Tcl_Alloc(strlen(substcmd)
				+ strlen(sres) + 1);
		    strcpy(newcmd, substcmd);
		    sprintf(newcmd + (int)(sptr - substcmd), "\"%s\"", sres);
		    strcat(newcmd, sptr + 2);
		    Tcl_Free(substcmd);
		    substcmd = newcmd;
		    sptr = substcmd;
		    break;

		case '0': case '1': case '2': case '3': case '4': case '5':
		    objidx = (int)(*(sptr + 1) - '0');
		    if ((objidx >= 0) && (objidx < objc))
		    {
		        newcmd = (char *)Tcl_Alloc(strlen(substcmd)
				+ strlen(Tcl_GetString(objv[objidx])));
		        strcpy(newcmd, substcmd);
			strcpy(newcmd + (int)(sptr - substcmd),
				Tcl_GetString(objv[objidx]));
			strcat(newcmd, sptr + 2);
			Tcl_Free(substcmd);
			substcmd = newcmd;
			sptr = substcmd;
		    }
		    else if (objidx >= objc)
		    {
		        newcmd = (char *)Tcl_Alloc(strlen(substcmd) + 1);
		        strcpy(newcmd, substcmd);
			strcpy(newcmd + (int)(sptr - substcmd), sptr + 2);
			Tcl_Free(substcmd);
			substcmd = newcmd;
			sptr = substcmd;
		    }
		    else sptr++;
		    break;

		case 'N':
		    llen = 1;
		    for (i = 1; i < objc; i++)
		       llen += (1 + strlen(Tcl_GetString(objv[i])));
		    newcmd = (char *)Tcl_Alloc(strlen(substcmd) + llen);
		    strcpy(newcmd, substcmd);
		    strcpy(newcmd + (int)(sptr - substcmd), "{");
		    for (i = 1; i < objc; i++) {
		       strcat(newcmd, Tcl_GetString(objv[i]));
		       if (i < (objc - 1))
			  strcat(newcmd, " ");
		    }
		    strcat(newcmd, "}");
		    strcat(newcmd, sptr + 2);
		    Tcl_Free(substcmd);
		    substcmd = newcmd;
		    sptr = substcmd;
		    break;

		case '%':
		    newcmd = (char *)Tcl_Alloc(strlen(substcmd) + 1);
		    strcpy(newcmd, substcmd);
		    strcpy(newcmd + (int)(sptr - substcmd), sptr + 1);
		    Tcl_Free(substcmd);
		    substcmd = newcmd;
		    sptr = substcmd;
		    break;

		default:
		    break;
	    }
	}

	/* Fprintf(stderr, "Substituted tag callback is \"%s\"\n", substcmd); */
	/* Flush(stderr); */

	Tcl_SaveResult(interp, &state);
	result = Tcl_Eval(interp, substcmd);
	if ((result == TCL_OK) && (reset == FALSE))
	    Tcl_RestoreResult(interp, &state);
	else
	    Tcl_DiscardResult(&state);

	Tcl_Free(substcmd);
    }
    return result;
}

/*--------------------------------------------------------------*/
/* Add a command tag callback					*/
/*--------------------------------------------------------------*/

int dnatcl_tag(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_HashEntry *entry;
    char *hstring;
    int new;

    if (objc != 2 && objc != 3)
	return TCL_ERROR;

    entry = Tcl_CreateHashEntry(&DNATagTable, Tcl_GetString(objv[1]), &new);
    if (entry == NULL) return TCL_ERROR;

    hstring = (char *)Tcl_GetHashValue(entry);
    if (objc == 2)
    {
	Tcl_SetResult(interp, hstring, NULL);
	return TCL_OK;
    }

    if (strlen(Tcl_GetString(objv[2])) == 0)
    {
	Tcl_DeleteHashEntry(entry);
    }
    else
    {
	hstring = Tcl_Strdup(Tcl_GetString(objv[2]));
	Tcl_SetHashValue(entry, hstring);
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Initialization						*/
/*--------------------------------------------------------------*/

int InitCallback(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *objv[])
{
   int option = 0;

   if (objc == 2) {
      if (!strcasecmp("space", Tcl_GetString(objv[1]))) {
         option = 1;
      }
      else {
         Tcl_SetResult(interp, "Syntax:  init [space]\n", NULL);
         return TCL_ERROR;
      }
   }
   else if (objc != 1) {
      Tcl_SetResult(interp, "Syntax:  init [space]\n", NULL);
      return TCL_ERROR;
   }

   destroy_world(option);
   reinitialize(option);

   tcl_printf(stdout, "\nStarting over. . .\n");
   draw_world();

   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Quit (Exit)							*/
/*--------------------------------------------------------------*/

int QuitCallback(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *objv[])
{
   exit(0);
   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Single simulation step					*/
/*--------------------------------------------------------------*/

int single_step(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *objv[])
{
   simulate_step();
   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Modify view for drawing					*/
/*--------------------------------------------------------------*/

int dnatcl_view(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   int result;
   double a, x, y, z;

   if (objc == 1) {
      // Return list of 4 components (angle, x, y, z)
      Tcl_Obj *lobj;

      lobj = Tcl_NewListObj(0, NULL);
      Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)viewangle[0]));
      Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)viewangle[1]));
      Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)viewangle[2]));
      Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)viewangle[3]));
      Tcl_SetObjResult(interp, lobj);
   }
   else if (objc == 5) {
      result = Tcl_GetDoubleFromObj(interp, objv[1], &a);
      if (result != TCL_OK) return result;
      result = Tcl_GetDoubleFromObj(interp, objv[2], &x);
      if (result != TCL_OK) return result;
      result = Tcl_GetDoubleFromObj(interp, objv[3], &y);
      if (result != TCL_OK) return result;
      result = Tcl_GetDoubleFromObj(interp, objv[4], &z);
      if (result != TCL_OK) return result;

      viewangle[0] = (float)a;
      viewangle[1] = (float)x;
      viewangle[2] = (float)y;
      viewangle[3] = (float)z;
   }
   else {
      Tcl_SetResult(interp, "Syntax:  view <angle> <x> <y> <z>\n", NULL);
      return TCL_ERROR;
   }
   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/

int dnatcl_zoom(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   int result;
   double z;

   if (objc == 1) {
      Tcl_SetObjResult(interp, Tcl_NewDoubleObj(viewzoom));
   }
   else if (objc == 2) {
      result = Tcl_GetDoubleFromObj(interp, objv[1], &z);
      if (result != TCL_OK) return result;

      viewzoom = (float)z;
   }
   else {
      Tcl_SetResult(interp, "Syntax:  zoom <value>\n", NULL);
      return TCL_ERROR;
   }
   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* Change verbose status					*/
/*--------------------------------------------------------------*/

int dnatcl_verbose(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   int value, result;

   if (objc != 1 && objc != 2)
      return TCL_ERROR;

   if (objc == 2)
   {
      result = Tcl_GetIntFromObj(interp, objv[1], &value);
      if (result != TCL_OK)
	 return result;

      verbose = value;
   }
   else {
     Tcl_SetObjResult(interp, Tcl_NewIntObj(verbose));
   }
   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* Zplane handling						*/
/*--------------------------------------------------------------*/

int dnatcl_zplane(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int value, result;
    char *token;

    if (objc != 1 && objc != 2)
	return TCL_ERROR;

    if (objc == 2)
    {
        token = Tcl_GetString(objv[1]);
	if (!strcasecmp(token, "off"))
	   zslice = (char)0;
	else if (!strcasecmp(token, "on"))
	   zslice = (char)1;
	else {
           result = Tcl_GetIntFromObj(interp, objv[1], &value);
	   if (result != TCL_OK) {
	      return result;
	   }
	   zplane = value;
	}
    }
    else {
       if (zslice == (char)0)
          Tcl_SetObjResult(interp, Tcl_NewStringObj("off", -1));
       else
          Tcl_SetObjResult(interp, Tcl_NewIntObj(zplane));
    }
    return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* We don't have a reverse lookup table for type and ruleset	*/
/* names, so we will have to loop through the hash table	*/
/* entries.							*/
/*--------------------------------------------------------------*/

int GetTypeName(Tcl_Interp *interp, Type *type, char **sreturn)
{
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   char *typename;
   Type *stype;

   he = Tcl_FirstHashEntry(&TypeHashTable, &hs);
   while (he != NULL) {
      stype = (Type *)Tcl_GetHashValue(he);
      if (stype == type) break;
      he = Tcl_NextHashEntry(&hs);
   }
   if (he == NULL) {	// Shouldn't happen
      Tcl_SetResult(interp, "Internal error:  Bad type in "
			"database!\n", NULL);
      return TCL_ERROR;
   }
   typename = (char *)Tcl_GetHashKey(&TypeHashTable, he);

   *sreturn = typename;
   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Likewise for rulesets. . .					*/
/*--------------------------------------------------------------*/

int GetRulesetName(Tcl_Interp *interp, DNARuleset *ruleset, char **sreturn)
{
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   char *rulesetname;
   DNARuleset *srule;

   he = Tcl_FirstHashEntry(&DNARuleSets, &hs);
   while (he != NULL) {
      srule = (DNARuleset *)Tcl_GetHashValue(he);
      if (srule == ruleset) break;
      he = Tcl_NextHashEntry(&hs);
   }
   if (he == NULL) {	// Shouldn't happen
      Tcl_SetResult(interp, "Internal error:  Bad ruleset in "
			"cell database!\n", NULL);
      return TCL_ERROR;
   }
   rulesetname = (char *)Tcl_GetHashKey(&DNARuleSets, he);
   *sreturn = rulesetname;
   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Handle data read-in through a command file using the		*/
/* commands "cell", "type", and "rule".  These can also be	*/
/* used on the command line to create and remove types, rules,	*/
/* and cells, as well as to list the contents of memory.	*/
/*--------------------------------------------------------------*/

int dnatcl_type(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Type *type;
   char *typename;
   Tcl_Obj *tlist;
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   int result;

   if (objc == 1) {
      /* Generate and return list of known types */ 

      tlist = Tcl_NewListObj(0, NULL);

      he = Tcl_FirstHashEntry(&TypeHashTable, &hs);
      while (he != NULL) {
	 typename = (char *)Tcl_GetHashKey(&TypeHashTable, he);
	 Tcl_ListObjAppendElement(interp, tlist, Tcl_NewStringObj(typename, -1));
	 he = Tcl_NextHashEntry(&hs);
      }
      Tcl_SetObjResult(interp, tlist);
   }
   else if (objc == 2) {

      /* Check type name;  does it exist? */
      he = Tcl_FindHashEntry(&TypeHashTable, Tcl_GetString(objv[1]));
      if (he == NULL) {
	 Tcl_SetResult(interp, "No such type name\n", NULL);
	 return TCL_ERROR;
      }

      /* Generate and return list of (r, g, b, a) for this type */ 

      type = (Type *)Tcl_GetHashValue(he);
      tlist = Tcl_NewListObj(0, NULL);

      Tcl_ListObjAppendElement(interp, tlist,
			Tcl_NewDoubleObj((double)type->r));
      Tcl_ListObjAppendElement(interp, tlist,
			Tcl_NewDoubleObj((double)type->g));
      Tcl_ListObjAppendElement(interp, tlist,
			Tcl_NewDoubleObj((double)type->b));
      Tcl_ListObjAppendElement(interp, tlist,
			Tcl_NewDoubleObj((double)type->a));
      Tcl_SetObjResult(interp, tlist);
   }

   else if (objc != 6 && objc != 3) {
      Tcl_SetResult(interp, "Need r g b a values with add option\n", NULL);
      return TCL_ERROR;
   } 
   else {
      double r, g, b, a;
      int l, newptr;
      Type *newtype;
      Tcl_Obj *lobj;

      if (objc == 6) {
         /* Get r, g, b, a from arguments 2, 3, 4, 5 */
	 result = Tcl_GetDoubleFromObj(interp, objv[2], &r);
	 if (result != TCL_OK) return result;
	 result = Tcl_GetDoubleFromObj(interp, objv[3], &g);
	 if (result != TCL_OK) return result;
	 result = Tcl_GetDoubleFromObj(interp, objv[4], &b);
	 if (result != TCL_OK) return result;
	 result = Tcl_GetDoubleFromObj(interp, objv[5], &a);
	 if (result != TCL_OK) return result;
     }
     else {
	 /* Argument 2 should be a list of length 4 containing r, g, b, a */
	 result = Tcl_ListObjLength(interp, objv[2], &l);
	 if (result != TCL_OK) return result;
	 if (l == 4) {
	    result = Tcl_ListObjIndex(interp, objv[2], 0, &lobj);
	    if (result != TCL_OK) return result;
	    Tcl_GetDoubleFromObj(interp, lobj, &r);
	    result = Tcl_ListObjIndex(interp, objv[2], 1, &lobj);
	    Tcl_GetDoubleFromObj(interp, lobj, &g);
	    result = Tcl_ListObjIndex(interp, objv[2], 2, &lobj);
	    Tcl_GetDoubleFromObj(interp, lobj, &b);
	    result = Tcl_ListObjIndex(interp, objv[2], 3, &lobj);
	    Tcl_GetDoubleFromObj(interp, lobj, &a);
         }
	 else {
	    Tcl_SetResult(interp, "Need list containing r, g, b, a components\n",
			NULL);
	    return TCL_ERROR;
	 }
      }

      /* First check if type exists */
      typename = Tcl_GetString(objv[1]);
      he = Tcl_CreateHashEntry(&TypeHashTable, typename, &newptr);
      if (newptr == 0) {
	 /* Modify existing record */
	 newtype = (Type *)Tcl_GetHashValue(he);
      }
      else {
	 /* Add new type */
	 newtype = (Type *)Tcl_Alloc(sizeof(Type));
	 Tcl_SetHashValue(he, newtype);

	 if (SpaceType == NULL) {
	    /* If we have not created a space type first, then it	*/
	    /* should be created automatically.				*/
	    if (strcmp(typename, "space") != 0) {
	       he = Tcl_CreateHashEntry(&TypeHashTable, "space", &newptr);
	       if (newptr == 1) {
	          type = (Type *)Tcl_Alloc(sizeof(Type));
		  type->r = 0.0;
		  type->g = 0.0;
		  type->b = 0.0;
		  type->a = 0.0;
		  type->idx = num_types++;
		  Tcl_SetHashValue(he, type);
	       }
	       else
	          type = (Type *)Tcl_GetHashValue(he);

	       SpaceType = type;
	    }
	    else
	       SpaceType = newtype;
	 }
	 newtype->idx = num_types++;
      }

      newtype->r = (float)r;
      newtype->g = (float)g;
      newtype->b = (float)b;
      newtype->a = (float)a;
   }

   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/

int dnatcl_rule(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Type *type;
   DNARuleset *ruleset;
   DNARule *rule, *lastrule;
   DNACond *cond, *nextcond;
   char *typename, *rulesetname, *resultname, *cmdname, *opname;
   int newptr, idx, l, m, i, j, number;
   Tcl_Obj *rlist, *rlist2, *clist, *tobj, *lobj, *cobj;
   Tcl_HashSearch hs;
   Tcl_HashEntry *he, *he2;
   int result;

   const char *subCmds[] = {"add", "remove", "list", NULL};
   enum SubIdx {ADD, REMOVE, LIST};

   if (objc == 1) {
      /* Generate and return list of known ruleset names */ 

      rlist = Tcl_NewListObj(0, NULL);

      he = Tcl_FirstHashEntry(&DNARuleSets, &hs);
      while (he != NULL) {
	 rulesetname = (char *)Tcl_GetHashKey(&DNARuleSets, he);
	 Tcl_ListObjAppendElement(interp, rlist, Tcl_NewStringObj(rulesetname, -1));
	 he = Tcl_NextHashEntry(&hs);
      }
      Tcl_SetObjResult(interp, rlist);
   }
   else if (objc == 2) {
      Tcl_SetResult(interp, "Syntax:  rule add|delete|list <options>\n", NULL);
      return TCL_ERROR;
   }
   else {
      if (Tcl_GetIndexFromObj(interp, objv[1], subCmds, "option", 0, &idx)
		!= TCL_OK)
	 return TCL_ERROR;

      /* Check ruleset name;  does it exist? */
      rulesetname = Tcl_GetString(objv[2]);
      he = Tcl_FindHashEntry(&DNARuleSets, rulesetname);

      switch (idx) {
	 case LIST:
	    if (he == NULL) {
	       Tcl_SetResult(interp, "No such ruleset.\n", NULL);
	       return TCL_ERROR;  
	    }
	    else {
	       ruleset = (DNARuleset *)Tcl_GetHashValue(he);

	       if (objc == 3) {
		  /* List all the types that have rules in this ruleset */
		  rlist = Tcl_NewListObj(0, NULL);

		  he = Tcl_FirstHashEntry(&(ruleset->DNARuleHash), &hs);
		  while (he != NULL) {
		     rule = Tcl_GetHashValue(he);
		     type = (Type *)Tcl_GetHashKey(&(ruleset->DNARuleHash), he);

		     // Get the type name for this rule type
		     result = GetTypeName(interp, type, &typename);
		     if (result != TCL_OK) return result;	// should not happen

		     Tcl_ListObjAppendElement(interp, rlist,
				Tcl_NewStringObj(typename, -1));

		     he = Tcl_NextHashEntry(&hs);
		  }
		  Tcl_SetObjResult(interp, rlist);
	       }
	       else if (objc == 4) {
		  typename = Tcl_GetString(objv[3]);
		  he = Tcl_FindHashEntry(&TypeHashTable, typename);
		  if (he == NULL) {
		     Tcl_SetResult(interp, "No such type.\n", NULL);
		     return TCL_ERROR;
		  }
		  else
		     type = (Type *)Tcl_GetHashValue(he);
		  rlist = Tcl_NewListObj(0, NULL);
		  he = Tcl_FindHashEntry(&(ruleset->DNARuleHash),
				(const char *)type);
		  if (he != NULL) {
		     /* List all rules for type in the ruleset */
		     rule = (DNARule *)Tcl_GetHashValue(he);
		     while (rule != NULL) {
			
			/* List first the result, then all conditions */
			result = GetTypeName(interp, rule->result, &resultname);
			if (result != TCL_OK) return result;

		        Tcl_ListObjAppendElement(interp, rlist,
				Tcl_NewStringObj(resultname, -1));
			clist = Tcl_NewListObj(0, NULL);
			cond = rule->condition;
			while (cond != NULL) {
			   result = GetTypeName(interp, cond->type, &typename);
			   if (result != TCL_OK) return result;
			   Tcl_ListObjAppendElement(interp, clist,
				Tcl_NewStringObj(typename, -1));
			   switch (cond->operator) {
			      case EQUAL:
			         Tcl_ListObjAppendElement(interp, clist,
					Tcl_NewStringObj("=", -1));
				 break;
			      case NOT_EQUAL:
			         Tcl_ListObjAppendElement(interp, clist,
					Tcl_NewStringObj("~", -1));
				 break;
			      case LESS_THAN:
			         Tcl_ListObjAppendElement(interp, clist,
					Tcl_NewStringObj("<", -1));
				 break;
			      case MORE_THAN:
			         Tcl_ListObjAppendElement(interp, clist,
					Tcl_NewStringObj(">", -1));
				 break;
			   }
			   Tcl_ListObjAppendElement(interp, clist, 
				Tcl_NewIntObj(cond->number));
	
			   cond = cond->next;
			}
			Tcl_ListObjAppendElement(interp, rlist, clist);
			rule = rule->next;
		     }
		  }
		  Tcl_SetObjResult(interp, rlist);
	       }
	       else {
		  Tcl_SetResult(interp, "Syntax:  rule list <ruleset> [<type>]\n",
				NULL);
		  return TCL_ERROR;
	       }
	    }
	    break;

	 case ADD:
	    if (objc == 4) {
	       /* "rule add <ruleset> <type>" is not legal syntax */
	       Tcl_SetResult(interp, "Syntax: rule add <ruleset> "
			"<type> {conditions}\n", NULL);
	       return TCL_ERROR;
	    }

	    /* Get typename from objv[3], make sure it is valid before	*/
	    /* we create the ruleset.					*/

	    if (objc != 3) {
	       typename = Tcl_GetString(objv[3]);
	       he2 = Tcl_FindHashEntry(&TypeHashTable, typename);
	       if (he2 == NULL) {
	          Tcl_SetResult(interp, "No such type.\n", NULL);
	          return TCL_ERROR;
	       }
	       else
	          type = (Type *)Tcl_GetHashValue(he2);
	    }

	    if (he == NULL) {
	       /* Create the ruleset */
	       ruleset = (DNARuleset *)Tcl_Alloc(sizeof(DNARuleset));
	       Tcl_InitHashTable(&(ruleset->DNARuleHash), TCL_ONE_WORD_KEYS);
	       he = Tcl_CreateHashEntry(&DNARuleSets, rulesetname, &newptr);
	       Tcl_SetHashValue(he, ruleset);
	    }
	    else
	       ruleset = Tcl_GetHashValue(he);

	    /* Syntax "rule add <rulename>" creates a NULL rule */
	    if (objc == 3)
	       return DNATagCallback(interp, objc, objv);

	    /* Create a new rule */
	    rule = (DNARule *)Tcl_Alloc(sizeof(DNARule));
	    rule->next = NULL;

	    he = Tcl_CreateHashEntry(&(ruleset->DNARuleHash),
			(const char *)type, &newptr);
	    if (newptr == 1) {
	       /* No rules for type "typename", so create one */
	       Tcl_SetHashValue(he, rule);
	       rule->condition = NULL;
	    }
	    else {
	       lastrule = (DNARule *)Tcl_GetHashValue(he);
	       while (lastrule->next != NULL)
		  lastrule = lastrule->next;
	       lastrule->next = rule;
	    }

	    /* Next argument should be a list {result {type op num ... }} */

	    result = Tcl_ListObjLength(interp, objv[4], &l);
	    if (result != TCL_OK) return result;
	    if (l & 1) {
	       Tcl_SetResult(interp, "Rule must be list {result condition ...}\n",
			NULL);
	       return TCL_ERROR;
	    }
	    for (i = 0; i < l; i += 2) {
	       result = Tcl_ListObjIndex(interp, objv[4], i, &tobj);
	       if (result != TCL_OK) return result;

	       /* Check for a valid type */
	       typename = Tcl_GetString(tobj);
	       he = Tcl_FindHashEntry(&TypeHashTable, typename);
	       if (he == NULL) {
		  Tcl_SetResult(interp, "No such type!\n", NULL);  
		  return TCL_ERROR;
	       }
	       else
		  type = (Type *)Tcl_GetHashValue(he);

	       rule->result = type;
	
	       result = Tcl_ListObjIndex(interp, objv[4], i + 1, &lobj);
	       if (result != TCL_OK) return result;
	       result = Tcl_ListObjLength(interp, lobj, &m);
	       if (m % 3 != 0) {
	          Tcl_SetResult(interp, "Rule condition must be list "
			"{type operator number ...}\n", NULL);
	          return TCL_ERROR;
	       }
	       rule->condition = NULL;
	       nextcond = NULL;

	       /* Create new condition */
	
	       for (j = 0; j < m; j += 3) {
	          result = Tcl_ListObjIndex(interp, lobj, j, &cobj);
	          if (result != TCL_OK) return result;
		  typename = Tcl_GetString(cobj);

	          he = Tcl_FindHashEntry(&TypeHashTable, typename);
	          if (he == NULL) {
		     Tcl_SetResult(interp, "No such type!\n", NULL);  
		     return TCL_ERROR;
	          }
	          else {
		     cond = (DNACond *)Tcl_Alloc(sizeof(DNACond));
		     cond->type = (Type *)Tcl_GetHashValue(he);
		  }

	          result = Tcl_ListObjIndex(interp, lobj, j + 1, &cobj);
	          if (result != TCL_OK) return result;
		  opname = Tcl_GetString(cobj);
		  switch (opname[0]) {
		     case '>':
			cond->operator = MORE_THAN; break;
		     case '<':
			cond->operator = LESS_THAN; break;
		     case '=':
			cond->operator = EQUAL; break;
		     case '~':
			cond->operator = NOT_EQUAL; break;
		     default:
			Tcl_SetResult(interp, "No such operator.\n", NULL);
			return TCL_ERROR;
		  }
	          result = Tcl_ListObjIndex(interp, lobj, j + 2, &cobj);
	          if (result != TCL_OK) return result;
		  result = Tcl_GetIntFromObj(interp, cobj, &number);
	          if (result != TCL_OK) return result;
		  cond->number = number;
		  cond->next = NULL;

		  if (nextcond == NULL)
		     rule->condition = cond;
		  else {
		     nextcond->next = cond;
		  }
		  nextcond = cond;
	       }
	       if (i + 2 < l) {
		  rule->next = (DNARule *)Tcl_Alloc(sizeof(DNARule));
		  rule = rule->next;
	       }
	       else
		  rule->next = NULL;
	    }
	    break;

	 case REMOVE:
	    if (he == NULL) {
	       Tcl_SetResult(interp, "No such ruleset.\n", NULL);
	       return TCL_ERROR;  
	    }
	    else {
	       Tcl_SetResult(interp, "Option \"rule remove\" not implemented.\n",
			NULL);
	       return TCL_ERROR;  
	    }
	    break;
      }
   }
   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/

int dnatcl_cell(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Cell *cell;
   Type *type;
   DNARuleset *ruleset;
   Point3D *cellpos, checkpos;
   Tcl_Obj *tlist, *clist, *lobj, *cobj;
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   char *rulename, *typename, *phash;
   int result, l;

   if (objc == 1) {
      /* Generate and return list of cell positions */ 

      tlist = Tcl_NewListObj(0, NULL);

      he = Tcl_FirstHashEntry(&CellHashTable, &hs);
      while (he != NULL) {
	 phash = Tcl_GetHashKey(&CellHashTable, he);
	 cellpos = (Point3D *)HashToPoint(phash);
	 clist = Tcl_NewListObj(0, NULL);
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->x));
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->y));
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->z));
	 Tcl_ListObjAppendElement(interp, tlist, clist);
	 he = Tcl_NextHashEntry(&hs);

	 if (verbose == 3) {
	    tcl_printf(stdout, "Hash \"%s\" position %d %d %d\n", 
			phash, cellpos->x, cellpos->y, cellpos->z);
	 }
      }
      Tcl_SetObjResult(interp, tlist);
      return DNATagCallback(interp, objc, objv);
   }
   else if (objc == 2) {
      result = Tcl_ListObjLength(interp, objv[1], &l);
      if (result != TCL_OK) return result;
      if (l != 3) {
	 Tcl_SetResult(interp, "Syntax: cell {<x> <y> <z>}\n", NULL);
	 return TCL_ERROR;
      }
      result = Tcl_ListObjIndex(interp, objv[1], 0, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.x));
      if (result != TCL_OK) return result;
      result = Tcl_ListObjIndex(interp, objv[1], 1, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.y));
      if (result != TCL_OK) return result;
      result = Tcl_ListObjIndex(interp, objv[1], 2, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.z));
      if (result != TCL_OK) return result;
   }
   else if (objc >= 4) {
      result = Tcl_GetIntFromObj(interp, objv[1], &(checkpos.x));
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, objv[2], &(checkpos.y));
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, objv[3], &(checkpos.z));
      if (result != TCL_OK) return result;
   }

   // Determine if this is a valid position.  Due to the use of euclidean
   // coordinates for a hex grid, not all positions are valid.

   if (((checkpos.y % 4) == 0) && (checkpos.x & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (1)\n", NULL);
      return TCL_ERROR;
   }
   if ((((checkpos.y + 2) % 4) == 0) && !(checkpos.x & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (2)\n", NULL);
      return TCL_ERROR;
   }
   if ((((checkpos.z + 1) % 2) == 0) && !(checkpos.y & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (3)\n", NULL);
      return TCL_ERROR;
   }
   if (((checkpos.z % 2) == 0) && (checkpos.y & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (4)\n", NULL);
      return TCL_ERROR;
   }
   if ((((checkpos.y + 1) % 2) == 0) && !(checkpos.z & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (5)\n", NULL);
      return TCL_ERROR;
   }
   if (((checkpos.y % 2) == 0) && (checkpos.z & 0x1)) {
      Tcl_SetResult(interp, "Invalid cell position, off hex grid (6)\n", NULL);
      return TCL_ERROR;
   }

   phash = PointToHash(&checkpos);

   if (verbose == 3) {
      tcl_printf(stdout, "Position %d %d %d hash is \"%s\"\n", 
	 checkpos.x, checkpos.y, checkpos.z, phash);
   }

   if (objc == 2 || objc == 4) {
      /* Generate and return type and ruleset information at a cell position */ 
      he = Tcl_FindHashEntry(&CellHashTable, phash);

      if (he == NULL) {
         Tcl_SetObjResult(interp, Tcl_NewStringObj("space", -1));
      }
      else {
	 cell = (Cell *)Tcl_GetHashValue(he);
         tlist = Tcl_NewListObj(0, NULL);

         result = GetTypeName(interp, cell->type, &typename);
         if (result != TCL_OK) return result;

         Tcl_ListObjAppendElement(interp, tlist, Tcl_NewStringObj(typename, -1));

	 result = GetRulesetName(interp, cell->ruleset, &rulename);
         if (result != TCL_OK) return result;
         Tcl_ListObjAppendElement(interp, tlist, Tcl_NewStringObj(rulename, -1));
      }
      Tcl_SetObjResult(interp, tlist);
   }
   else if (objc == 5) {
      /* Type name must be "space" for use with only 5 arguments */
      typename = Tcl_GetString(objv[4]);
      if (strcasecmp(typename, "space") != 0) {
         Tcl_SetResult(interp, "Type other than \"space\" must have ruleset", NULL);
	 return TCL_ERROR;
      }
      AddChangeEvent(phash, SpaceType, NULL);
   }
   else if (objc == 6) {
      typename = Tcl_GetString(objv[4]);
      rulename = Tcl_GetString(objv[5]);

      he = Tcl_FindHashEntry(&TypeHashTable, typename);
      if (he == NULL) {
	 Tcl_SetResult(interp, "No such type name\n", NULL);
	 return TCL_ERROR;
      }
      else
	 type = (Type *)Tcl_GetHashValue(he);

      he = Tcl_FindHashEntry(&DNARuleSets, rulename);
      if (he == NULL) {
	 Tcl_SetResult(interp, "No such ruleset name\n", NULL);
	 return TCL_ERROR;
      }
      else
	 ruleset = (DNARuleset *)Tcl_GetHashValue(he);

      AddChangeEvent(phash, type, ruleset);
   }
   else {
      Tcl_SetResult(interp, "Syntax:  cell [x y z [type ruleset]]\n", NULL);
      return TCL_ERROR;
   }

   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/

int dnatcl_pending(ClientData clientData,
        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Cell *cell;
   Type *type;
   DNARuleset *ruleset;
   Point3D *cellpos, checkpos, *eventpos;
   Tcl_Obj *tlist, *clist, *lobj, *cobj;
   Tcl_HashSearch hs;
   Tcl_HashEntry *he;
   char *rulename, *typename, *phash;
   int result, l;
   ChangeRecord *event;

   if (objc == 1) {
      /* Generate and return list of positions of change events */ 

      tlist = Tcl_NewListObj(0, NULL);

      event = ChangeEventStack;
      while (event != NULL) {
	 cellpos = HashToPoint(event->phash);
	 clist = Tcl_NewListObj(0, NULL);
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->x));
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->y));
	 Tcl_ListObjAppendElement(interp, clist, Tcl_NewIntObj(cellpos->z));
	 Tcl_ListObjAppendElement(interp, tlist, clist);

	 if (verbose == 3) {
	    tcl_printf(stdout, "Hash \"%s\" position %d %d %d\n", 
			event->phash, cellpos->x, cellpos->y, cellpos->z);
	 }
	 event = event->next;
      }
      Tcl_SetObjResult(interp, tlist);
      return DNATagCallback(interp, objc, objv);
   }
   else if (objc == 2) {
      result = Tcl_ListObjLength(interp, objv[1], &l);
      if (result != TCL_OK) return result;
      if (l != 3) {
	 Tcl_SetResult(interp, "Syntax: cell {<x> <y> <z>}\n", NULL);
	 return TCL_ERROR;
      }
      result = Tcl_ListObjIndex(interp, objv[1], 0, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.x));
      if (result != TCL_OK) return result;
      result = Tcl_ListObjIndex(interp, objv[1], 1, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.y));
      if (result != TCL_OK) return result;
      result = Tcl_ListObjIndex(interp, objv[1], 2, &cobj);
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, cobj, &(checkpos.z));
      if (result != TCL_OK) return result;
   }
   else if (objc >= 4) {
      result = Tcl_GetIntFromObj(interp, objv[1], &(checkpos.x));
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, objv[2], &(checkpos.y));
      if (result != TCL_OK) return result;
      result = Tcl_GetIntFromObj(interp, objv[3], &(checkpos.z));
      if (result != TCL_OK) return result;
   }
   phash = PointToHash(&checkpos);

   if (objc == 2 || objc == 4) {
      /* Generate and return type and ruleset information at a cell position */ 
      he = Tcl_FindHashEntry(&CellHashTable, phash);
      tlist = Tcl_NewListObj(0, NULL);

      event = ChangeEventStack;
      while (event != NULL) {
	 eventpos = HashToPoint(event->phash);
	 if (eventpos->x == checkpos.x && eventpos->y == checkpos.y &&
			eventpos->z == checkpos.z) {
	    /* Return list size 2 containing the event type and ruleset */
	    tlist = Tcl_NewListObj(0, NULL);
	    result = GetTypeName(interp, event->type, &typename);
	    if (result != TCL_OK) return result;
            Tcl_ListObjAppendElement(interp, tlist, Tcl_NewStringObj(typename, -1));

	    result = GetRulesetName(interp, event->ruleset, &rulename);
            if (result != TCL_OK) return result;
            Tcl_ListObjAppendElement(interp, tlist, Tcl_NewStringObj(rulename, -1));
	    Tcl_SetObjResult(interp, tlist);
	    break;
	 }
	 event = event->next;
      }
   }
   else {
      Tcl_SetResult(interp, "Syntax:  pending [{x y z}]\n", NULL);
      return TCL_ERROR;
   }
   return DNATagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* Initialization procedure for Tcl/Tk				*/
/*--------------------------------------------------------------*/

int
Dna_Init(Tcl_Interp *interp)
{
   int cmdidx;
   Tk_Window tktop;
   char *tmp_s;

   /* Interpreter sanity checks */
   if (interp == NULL) return TCL_ERROR;

   /* Remember the interpreter */
   dnainterp = interp;

   if (Tcl_InitStubs(interp, "8.1", 0) == NULL) return TCL_ERROR;
   
   /* Create the start command */

   tktop = Tk_MainWindow(interp);

   /* Create all of the commands (except "simple") */

   for (cmdidx = 0; dna_commands[cmdidx].func != NULL; cmdidx++) {
      Tcl_CreateObjCommand(interp, dna_commands[cmdidx].cmdstr, 
		(Tcl_ObjCmdProc *)dna_commands[cmdidx].func,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);
   }

   /* Command which creates a "simple" window.		*/

   Tcl_CreateObjCommand(interp, "simple",
		(Tcl_ObjCmdProc *)Tk_SimpleObjCmd,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);

   Tcl_Eval(interp, "lappend auto_path .");
   Tcl_PkgProvide(interp, "DNA", "0.1");

   /* Initialize the console interpreter, if there is one. */

   if ((consoleinterp = Tcl_GetMaster(interp)) == NULL)
      consoleinterp = interp;

   /* Initialize the command tag table */

   Tcl_InitHashTable(&DNATagTable, TCL_STRING_KEYS);

   return TCL_OK;
}

/*------------------------------------------------------*/
/* Command functions					*/
/*------------------------------------------------------*/

int dna_start(ClientData clientData, Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
   GUI_init(interp, objc, objv);
   return DNATagCallback(interp, objc, objv);
}

/*------------------------------------------------------*/
/*------------------------------------------------------*/


